@@ -25,7 +25,7 @@
 // OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
 // IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 // DEALINGS IN THE SOFTWARE.
-use crate::opentelemetry::PreSampledTracer;
+use crate::opentelemetry::{OtelData, PreSampledTracer};
 use opentelemetry::{
     trace::{self as otel, noop, TraceContextExt},
     Context as OtelContext, Key, KeyValue,
@@ -47,7 +47,7 @@ use crate::{SampleDecision, TraceContext};
 
 #[derive(Default)]
 struct TraceCache {
-    spans: VecDeque<otel::SpanBuilder>,
+    spans: VecDeque<OtelData>,
 }
 
 impl TraceCache {
@@ -55,13 +55,15 @@ impl TraceCache {
     where
         T: otel::Tracer + PreSampledTracer + 'static,
     {
-        while let Some(span) = self.spans.pop_front() {
-            span.start(tracer);
+        let trace_spans = std::mem::replace(&mut self.spans, Default::default());
+
+        for OtelData { builder, parent_cx } in trace_spans {
+            builder.start_with_context(tracer, &parent_cx);
         }
     }
 
     fn clear(&mut self) {
-        self.spans.clear();
+        drop(std::mem::replace(&mut self.spans, Default::default()));
     }
 }
 
@@ -119,11 +121,7 @@ where
 //
 // See https://github.com/tokio-rs/tracing/blob/4dad420ee1d4607bad79270c1520673fa6266a3d/tracing-error/src/layer.rs
 pub(crate) struct WithContext(
-    fn(
-        &tracing::Dispatch,
-        &span::Id,
-        f: &mut dyn FnMut(&mut otel::SpanBuilder, &dyn PreSampledTracer),
-    ),
+    fn(&tracing::Dispatch, &span::Id, f: &mut dyn FnMut(&mut OtelData, &dyn PreSampledTracer)),
 );
 
 impl WithContext {
@@ -133,7 +131,7 @@ impl WithContext {
         &self,
         dispatch: &'a tracing::Dispatch,
         id: &span::Id,
-        mut f: impl FnMut(&mut otel::SpanBuilder, &dyn PreSampledTracer),
+        mut f: impl FnMut(&mut OtelData, &dyn PreSampledTracer),
     ) {
         (self.0)(dispatch, id, &mut f)
     }
@@ -410,7 +408,7 @@ where
             let span = ctx.span(parent).expect("Span not found, this is a bug");
             let mut extensions = span.extensions_mut();
             extensions
-                .get_mut::<otel::SpanBuilder>()
+                .get_mut::<OtelData>()
                 .map(|builder| self.tracer.sampled_context(builder))
                 .unwrap_or_default()
         // Else if the span is inferred from context, look up any available current span.
@@ -419,7 +417,7 @@ where
                 .and_then(|span| {
                     let mut extensions = span.extensions_mut();
                     extensions
-                        .get_mut::<otel::SpanBuilder>()
+                        .get_mut::<OtelData>()
                         .map(|builder| self.tracer.sampled_context(builder))
                 })
                 .unwrap_or_else(OtelContext::current)
@@ -432,7 +430,7 @@ where
     fn get_context(
         dispatch: &tracing::Dispatch,
         id: &span::Id,
-        f: &mut dyn FnMut(&mut otel::SpanBuilder, &dyn PreSampledTracer),
+        f: &mut dyn FnMut(&mut OtelData, &dyn PreSampledTracer),
     ) {
         let subscriber = dispatch
             .downcast_ref::<S>()
@@ -445,7 +443,7 @@ where
             .expect("layer should downcast to expected type; this is a bug!");
 
         let mut extensions = span.extensions_mut();
-        if let Some(builder) = extensions.get_mut::<otel::SpanBuilder>() {
+        if let Some(builder) = extensions.get_mut::<OtelData>() {
             f(builder, &layer.tracer);
         }
     }
@@ -468,16 +466,16 @@ where
             extensions.insert(Timings::new());
         }
 
+        let parent_cx = self.parent_context(attrs, &ctx);
         let mut builder = self
             .tracer
             .span_builder(attrs.metadata().name())
             .with_start_time(SystemTime::now())
-            .with_parent_context(self.parent_context(attrs, &ctx))
             // Eagerly assign span id so children have stable parent id
             .with_span_id(self.tracer.new_span_id());
 
         // Record new trace id if there is no active parent span
-        if !builder.parent_context.has_active_span() {
+        if !parent_cx.has_active_span() {
             builder.trace_id = Some(self.tracer.new_trace_id());
         }
 
@@ -500,7 +498,7 @@ where
         }
 
         attrs.record(&mut SpanAttributeVisitor(&mut builder));
-        extensions.insert(builder);
+        extensions.insert(OtelData { builder, parent_cx });
     }
 
     fn on_enter(&self, id: &span::Id, ctx: Context<'_, S>) {
@@ -539,37 +537,37 @@ where
     fn on_record(&self, id: &Id, values: &Record<'_>, ctx: Context<'_, S>) {
         let span = ctx.span(id).expect("Span not found, this is a bug");
         let mut extensions = span.extensions_mut();
-        if let Some(builder) = extensions.get_mut::<otel::SpanBuilder>() {
-            values.record(&mut SpanAttributeVisitor(builder));
+        if let Some(data) = extensions.get_mut::<OtelData>() {
+            values.record(&mut SpanAttributeVisitor(&mut data.builder));
         }
     }
 
     fn on_follows_from(&self, id: &Id, follows: &Id, ctx: Context<S>) {
         let span = ctx.span(id).expect("Span not found, this is a bug");
         let mut extensions = span.extensions_mut();
-        let builder = extensions
-            .get_mut::<otel::SpanBuilder>()
-            .expect("Missing SpanBuilder span extensions");
+        let data = extensions
+            .get_mut::<OtelData>()
+            .expect("Missing otel data span extensions");
 
         let follows_span = ctx
             .span(follows)
             .expect("Span to follow not found, this is a bug");
         let mut follows_extensions = follows_span.extensions_mut();
-        let follows_builder = follows_extensions
-            .get_mut::<otel::SpanBuilder>()
-            .expect("Missing SpanBuilder span extensions");
+        let follows_data = follows_extensions
+            .get_mut::<OtelData>()
+            .expect("Missing otel data span extensions");
 
         let follows_context = self
             .tracer
-            .sampled_context(follows_builder)
+            .sampled_context(follows_data)
             .span()
             .span_context()
             .clone();
         let follows_link = otel::Link::new(follows_context, Vec::new());
-        if let Some(ref mut links) = builder.links {
+        if let Some(ref mut links) = data.builder.links {
             links.push(follows_link);
         } else {
-            builder.links = Some(vec![follows_link]);
+            data.builder.links = Some(vec![follows_link]);
         }
     }
 
@@ -604,7 +602,7 @@ where
             event.record(&mut SpanEventVisitor(&mut otel_event));
 
             let mut extensions = span.extensions_mut();
-            if let Some(builder) = extensions.get_mut::<otel::SpanBuilder>() {
+            if let Some(OtelData { builder, .. }) = extensions.get_mut::<OtelData>() {
                 if builder.status_code.is_none() && *meta.level() == tracing_core::Level::ERROR {
                     builder.status_code = Some(otel::StatusCode::Error);
                 }
@@ -624,7 +622,11 @@ where
     fn on_close(&self, id: span::Id, ctx: Context<'_, S>) {
         let span = ctx.span(&id).expect("Span not found, this is a bug");
         let mut extensions = span.extensions_mut();
-        if let Some(mut builder) = extensions.remove::<otel::SpanBuilder>() {
+        if let Some(OtelData {
+            mut builder,
+            parent_cx,
+        }) = extensions.remove::<OtelData>()
+        {
             if self.tracked_inactivity {
                 // Append busy/idle timings when enabled.
                 if let Some(timings) = extensions.get_mut::<Timings>() {
@@ -660,7 +662,7 @@ where
                     .get_mut::<TraceCache>()
                     .expect("Cache not found, this is a bug");
 
-                cache.spans.push_back(builder);
+                cache.spans.push_back(OtelData { builder, parent_cx });
 
                 // Now, if this is the top level span, see if we can flush.
                 if trace_context.parent_id.is_none() {
@@ -672,7 +674,7 @@ where
                 }
             } else {
                 // build and start span, drop span to export
-                builder.start(&self.tracer);
+                builder.start_with_context(&self.tracer, &parent_cx);
             }
         }
     }
@@ -709,6 +711,7 @@ impl Timings {
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::opentelemetry::OtelData;
     use opentelemetry::trace::{noop, SpanKind, TraceFlags};
     use std::borrow::Cow;
     use std::sync::{Arc, Mutex};
@@ -716,17 +719,14 @@ mod tests {
     use tracing_subscriber::prelude::*;
 
     #[derive(Debug, Clone)]
-    struct TestTracer(Arc<Mutex<Option<otel::SpanBuilder>>>);
+    struct TestTracer(Arc<Mutex<Option<OtelData>>>);
     impl otel::Tracer for TestTracer {
         type Span = noop::NoopSpan;
-        fn invalid(&self) -> Self::Span {
-            noop::NoopSpan::new()
-        }
-        fn start_with_context<T>(&self, _name: T, _context: OtelContext) -> Self::Span
+        fn start_with_context<T>(&self, _name: T, _context: &OtelContext) -> Self::Span
         where
             T: Into<Cow<'static, str>>,
         {
-            self.invalid()
+            noop::NoopSpan::new()
         }
         fn span_builder<T>(&self, name: T) -> otel::SpanBuilder
         where
@@ -734,28 +734,42 @@ mod tests {
         {
             otel::SpanBuilder::from_name(name)
         }
-        fn build(&self, builder: otel::SpanBuilder) -> Self::Span {
-            *self.0.lock().unwrap() = Some(builder);
-            self.invalid()
+        fn build_with_context(
+            &self,
+            builder: otel::SpanBuilder,
+            parent_cx: &OtelContext,
+        ) -> Self::Span {
+            *self.0.lock().unwrap() = Some(OtelData {
+                builder,
+                parent_cx: parent_cx.clone(),
+            });
+            noop::NoopSpan::new()
         }
     }
 
     impl PreSampledTracer for TestTracer {
-        fn sampled_context(&self, _builder: &mut otel::SpanBuilder) -> OtelContext {
+        fn sampled_context(&self, _builder: &mut OtelData) -> OtelContext {
             OtelContext::new()
         }
         fn new_trace_id(&self) -> otel::TraceId {
-            otel::TraceId::invalid()
+            otel::TraceId::INVALID
         }
         fn new_span_id(&self) -> otel::SpanId {
-            otel::SpanId::invalid()
+            otel::SpanId::INVALID
         }
     }
 
     #[derive(Debug, Clone)]
     struct TestSpan(otel::SpanContext);
     impl otel::Span for TestSpan {
-        fn add_event_with_timestamp(&mut self, _: String, _: SystemTime, _: Vec<KeyValue>) {}
+        fn add_event_with_timestamp<T: Into<Cow<'static, str>>>(
+            &mut self,
+            _: T,
+            _: SystemTime,
+            _: Vec<KeyValue>,
+        ) {
+        }
+
         fn span_context(&self) -> &otel::SpanContext {
             &self.0
         }
@@ -764,7 +778,7 @@ mod tests {
         }
         fn set_attribute(&mut self, _attribute: KeyValue) {}
         fn set_status(&mut self, _code: otel::StatusCode, _message: String) {}
-        fn update_name(&mut self, _new_name: String) {}
+        fn update_name<T: Into<Cow<'static, str>>>(&mut self, _new_name: T) {}
         fn end_with_timestamp(&mut self, _timestamp: SystemTime) {}
     }
 
@@ -778,7 +792,13 @@ mod tests {
             tracing::debug_span!("static_name", otel.name = dynamic_name.as_str());
         });
 
-        let recorded_name = tracer.0.lock().unwrap().as_ref().map(|b| b.name.clone());
+        let recorded_name = tracer
+            .0
+            .lock()
+            .unwrap()
+            .as_ref()
+            .map(|b| b.builder.name.clone());
+
         assert_eq!(recorded_name, Some(dynamic_name.into()))
     }
 
@@ -791,7 +811,16 @@ mod tests {
             tracing::debug_span!("request", otel.kind = %SpanKind::Server);
         });
 
-        let recorded_kind = tracer.0.lock().unwrap().as_ref().unwrap().span_kind.clone();
+        let recorded_kind = tracer
+            .0
+            .lock()
+            .unwrap()
+            .as_ref()
+            .unwrap()
+            .builder
+            .span_kind
+            .clone();
+
         assert_eq!(recorded_kind, Some(otel::SpanKind::Server))
     }
 
@@ -803,7 +832,15 @@ mod tests {
         tracing::subscriber::with_default(subscriber, || {
             tracing::debug_span!("request", otel.status_code = ?otel::StatusCode::Ok);
         });
-        let recorded_status_code = tracer.0.lock().unwrap().as_ref().unwrap().status_code;
+        let recorded_status_code = tracer
+            .0
+            .lock()
+            .unwrap()
+            .as_ref()
+            .unwrap()
+            .builder
+            .status_code;
+
         assert_eq!(recorded_status_code, Some(otel::StatusCode::Ok))
     }
 
@@ -824,6 +861,7 @@ mod tests {
             .unwrap()
             .as_ref()
             .unwrap()
+            .builder
             .status_message
             .clone();
 
@@ -834,10 +872,10 @@ mod tests {
     fn trace_id_from_existing_context() {
         let tracer = TestTracer(Arc::new(Mutex::new(None)));
         let subscriber = tracing_subscriber::registry().with(layer().with_tracer(tracer.clone()));
-        let trace_id = otel::TraceId::from_u128(42);
+        let trace_id = otel::TraceId::from(42u128.to_be_bytes());
         let existing_cx = OtelContext::current_with_span(TestSpan(otel::SpanContext::new(
             trace_id,
-            otel::SpanId::from_u64(1),
+            otel::SpanId::from(1u64.to_be_bytes()),
             TraceFlags::default(),
             false,
             Default::default(),
@@ -854,10 +892,42 @@ mod tests {
             .unwrap()
             .as_ref()
             .unwrap()
-            .parent_context
+            .parent_cx
             .span()
             .span_context()
             .trace_id();
         assert_eq!(recorded_trace_id, trace_id)
     }
+
+    #[test]
+    fn includes_timings() {
+        let tracer = TestTracer(Arc::new(Mutex::new(None)));
+        let subscriber = tracing_subscriber::registry().with(
+            layer()
+                .with_tracer(tracer.clone())
+                .with_tracked_inactivity(true),
+        );
+
+        tracing::subscriber::with_default(subscriber, || {
+            tracing::debug_span!("request");
+        });
+
+        let attributes = tracer
+            .0
+            .lock()
+            .unwrap()
+            .as_ref()
+            .unwrap()
+            .builder
+            .attributes
+            .as_ref()
+            .unwrap()
+            .clone();
+        let keys = attributes
+            .iter()
+            .map(|attr| attr.key.as_str())
+            .collect::<Vec<&str>>();
+        assert!(keys.contains(&"idle_ns"));
+        assert!(keys.contains(&"busy_ns"));
+    }
 }
